// ESP 8266 support

#ifdef ESP

// ESP soft serial config:
#define ESP_SERIAL_RX_PIN   A2
#define ESP_SERIAL_TX_PIN   A3

// AT command strings stored in progmem
prog_char PROGMEM ESP_01[] = "AT+RST";
prog_char PROGMEM ESP_02[] = "AT+CWMODE=1";
prog_char PROGMEM ESP_03[] = "AT+CWJAP=\"";
prog_char PROGMEM ESP_04[] = "AT+CIFSR";
prog_char PROGMEM ESP_05[] = "AT+CIPMUX=1";
prog_char PROGMEM ESP_06[] = "AT+CIPSERVER=1,514";
prog_char PROGMEM ESP_07[] = "AT+CIPSTART=2,";
prog_char PROGMEM ESP_08[] = ESP_SSID;
prog_char PROGMEM ESP_09[] = ESP_PASS;
prog_char PROGMEM ESP_10[] = ESP_BROADCAST;
prog_char PROGMEM ESP_11[] = "AT+CIPSEND=2,64";
prog_char PROGMEM ESP_12[] = "\"UDP\",\"";
prog_char PROGMEM ESP_13[] = "\",514";
PROGMEM const char *ESPText_tabel[]={ESP_01,ESP_02,ESP_03,ESP_04,ESP_05,ESP_06,ESP_07,ESP_08,ESP_09,ESP_10,ESP_11,ESP_12,ESP_13};

#define ESP_PAYLOAD_NODO           0

struct ESPPayloadStruct
{
  byte Source;
  byte Destination;
  byte ID;
  byte Size;
  byte Sequence;
  byte Flags;
  byte Future1;
  byte Future2;
  byte Data[24];
};

byte ESP_Sequence=0;

boolean esp_init()
{

  char cmd[50];

  SoftwareSerial_init(ESP_SERIAL_RX_PIN,ESP_SERIAL_TX_PIN);         // RX, TX

  softSerialSend(ProgmemString(ESP_01),1000);
  delay(3000);
  softSerialSend(ProgmemString(ESP_02),1000);

  strcpy(cmd,ProgmemString(ESP_03));
  strcat(cmd,ProgmemString(ESP_08));
  strcat(cmd,"\",\"");
  strcat(cmd,ProgmemString(ESP_09));
  strcat(cmd,"\"");
  softSerialSend(cmd, 5000);

  softSerialSend(ProgmemString(ESP_04),500);
  softSerialSend(ProgmemString(ESP_05),500);
  softSerialSend(ProgmemString(ESP_06),500);

  strcpy(cmd,ProgmemString(ESP_07));
  strcat(cmd,ProgmemString(ESP_12));
  strcat(cmd,ESP_BROADCAST);
  strcat(cmd,ProgmemString(ESP_13));
  softSerialSend(cmd,500);

}

byte esp_event(void)
{
  byte returnvalue=0;
  
      if (SoftwareSerial_available())
        {
          delay(80);

          //bericht voorbeeld:
          //+IPD,0,28:1EFF000E6F00002F013A00D20100001E0027124D0A038C308C303B0802308C01
          //OK
          //
          // inclusief CRLF in totaal 82 bytes (CRLF,start 10 bytes, 64 bytes data, CRLF, OK, CRLF)
          // dit duurt ong 85 mSec bij 9600 baud
          // We starten met inlezen na de : , en stoppen bij CRLF
          // dus alleen die 64 bytes data wordt opgepakt

          #ifdef DEBUG
              Serial.println(SoftwareSerial_available());
          #endif
          char packetBuffer[65]; //buffer to hold incoming packet,
          byte x = 0;
          char c = 0;
          boolean foundsign = false;
          while (SoftwareSerial_available() && x < 64)
            {
              c = SoftwareSerial_read();
              if (foundsign)
                {
                  packetBuffer[x] = c;
                  x++;
                }
              if (c == ':') foundsign = true;
              if (foundsign && c == 13) break;
            }
          packetBuffer[x] = 0;

          if(SoftwareSerial_available())
            SoftwareSerial_flush();

          #ifdef DEBUG
              Serial.print("ESP<");
              Serial.println(packetBuffer);
          #endif

          if (foundsign)
            {
              struct ESPPayloadStruct ESPPayload;
              for (byte x=0; x < sizeof(struct ESPPayloadStruct); x++)
                {
                  byte b1 = packetBuffer[x*2] - '0';
                  byte b2 = packetBuffer[x*2+1] -'0';
                  if (b1 > 16) b1=b1-7;
                  if (b2 > 16) b2=b2-7;
                  b1 = b1*16 + b2;
                  memcpy((byte*)&ESPPayload+x,&b1,1);
                }
                
              struct NodoEventStruct TempEvent;
              memcpy((byte*)&TempEvent, (byte*)&ESPPayload+8,sizeof(struct NodoEventStruct));
                
              if(Checksum(&TempEvent))
                {
                  if(TempEvent.Type == NODO_TYPE_EVENT && TempEvent.Command==EVENT_USEREVENT && TempEvent.Par1==EVENT_PAR1)
                    {
                      returnvalue=TempEvent.Par2;
                    }
                  
                  if(TempEvent.Type == NODO_TYPE_COMMAND && TempEvent.Command==CMD_CLOCK_TIME)
                    {
                      byte Hour =((TempEvent.Par2>>12)&0xf)*10 + ((TempEvent.Par2>>8)&0xf);
                      byte Minutes =((TempEvent.Par2>>4 )&0xf)*10 + ((TempEvent.Par2   )&0xf);
                      setTime(Hour,Minutes,second(),day(),month(), year());
                      RTC.set(now());
                      #ifdef DEBUG
                        Serial.print("ESP CLOCK MSG ");
                        Serial.print(Hour);
                        Serial.print(':');
                        Serial.println(Minutes);
                      #endif
                    }

                  #ifdef DEBUG
                    Serial.println("ESP MSG OK");
                  #endif
                }
              else
                { 
                  #ifdef DEBUG
                    Serial.println("CRC!");
                  #endif
                }
            }
        } 
  return returnvalue;
}
  
boolean softSerialSend(char *cmd, int intdelay)
{
  char currentchar=0;
  char lastchar=0;
  unsigned long timer=millis()+ intdelay;

  SoftwareSerial_flush();

  #ifdef DEBUG
    Serial.println(cmd);
  #endif

  SoftwareSerial_println(cmd);

  while (millis() < timer)
    {
      while (SoftwareSerial_available())
        {
          currentchar = SoftwareSerial_read();
          #ifdef DEBUG
            Serial.write(currentchar);
          #endif
          if ( (lastchar == 'O' && currentchar == 'K') || currentchar == '>')
            {
              if(SoftwareSerial_available())
                SoftwareSerial_flush();
              #ifdef DEBUG
                Serial.println();
                Serial.print("OK-");
                Serial.println(millis()-(timer-intdelay));
              #endif
              return true;
            }
          lastchar = currentchar;
        }
    }
  #ifdef DEBUG
    Serial.println("NOK");
  #endif
  return false;
}

char* byte2strhex(byte x)
{
  static char hexString[3];
  hexString[2]=0;

  hexString[0] = (x >> 4) + 0x30;
  if (hexString[0] > 0x39) hexString[0] +=7;

  hexString[1] = (x & 0x0f) + 0x30;
  if (hexString[1] > 0x39) hexString[1] +=7;
  return (char*)&hexString;
}


/*
  Software serial gebaseerd op originele softwareserial lib
  Waarom deze code?
   - Include lib vanuit plugin file werkt niet
   - Code vereenvoudig en gereduceerd ivm Nodo plugin gebruik op Atmel 328 platform
*/

#define _SS_MAX_RX_BUFF         96 // RX buffer size

#if F_CPU == 16000000
  // settings for 16 MHz CPU, 9600 BAUD
  // Intrabit timing slightly adjusted, because program code is reduced
  #define RX_DELAY_CENTERING      114
  #define RX_DELAY_INTRABIT       236 // original 236
  #define RX_DELAY_STOPBIT        236
  #define TX_DELAY                233
  #define XMIT_START_ADJUSTMENT     5
#else
  // settings for 8 MHz CPU, 9600 BAUD
  // Intrabit timing slightly adjusted, because program code is reduced
  #define RX_DELAY_CENTERING      50
  #define RX_DELAY_INTRABIT       114 // original 236
  #define RX_DELAY_STOPBIT        114
  #define TX_DELAY                112
  #define XMIT_START_ADJUSTMENT     4
#endif

uint8_t _receivePin;
uint8_t _receiveBitMask;
volatile uint8_t *_receivePortRegister;
char _receive_buffer[_SS_MAX_RX_BUFF]; 
volatile uint8_t _receive_buffer_tail = 0;
volatile uint8_t _receive_buffer_head = 0;

uint8_t _transmitBitMask;
volatile uint8_t *_transmitPortRegister;

void SoftwareSerial_print(char* string)
{
  byte x=0;
  while (string[x] != 0)
  {
    SoftwareSerial_write(string[x]);
    x++;
  }
}

void SoftwareSerial_println(char* string)
{
  SoftwareSerial_print(string);
  SoftwareSerial_write(13);
  SoftwareSerial_write(10);
}

void SoftwareSerial_init(byte receivePin, byte transmitPin)
{
  SoftwareSerial_setTX(transmitPin);
  SoftwareSerial_setRX(receivePin);
  if (digitalPinToPCICR(_receivePin))
  {
    *digitalPinToPCICR(_receivePin) |= _BV(digitalPinToPCICRbit(_receivePin));
    *digitalPinToPCMSK(_receivePin) |= _BV(digitalPinToPCMSKbit(_receivePin));
  }
  SoftwareSerial_tunedDelay(TX_DELAY); // if we were low this establishes the end
}

inline void SoftwareSerial_tunedDelay(uint16_t delay) { 
  uint8_t tmp=0;

  asm volatile("sbiw    %0, 0x01 \n\t"
    "ldi %1, 0xFF \n\t"
    "cpi %A0, 0xFF \n\t"
    "cpc %B0, %1 \n\t"
    "brne .-10 \n\t"
    : "+w" (delay), "+a" (tmp)
    : "0" (delay)
    );
}

void SoftwareSerial_recv()
{
  uint8_t d = 0;

  // If RX line is high, then we don't see any start bit
  // so interrupt is probably not for us
  if (!SoftwareSerial_rx_pin_read())
  {
    // Wait approximately 1/2 of a bit width to "center" the sample
    SoftwareSerial_tunedDelay(RX_DELAY_CENTERING);

    // Read each of the 8 bits
    for (uint8_t i=0x1; i; i <<= 1)
    {
      SoftwareSerial_tunedDelay(RX_DELAY_INTRABIT);
      uint8_t noti = ~i;
      if (SoftwareSerial_rx_pin_read())
        d |= i;
      else // else clause added to ensure function timing is ~balanced
        d &= noti;
    }

    // skip the stop bit
    SoftwareSerial_tunedDelay(RX_DELAY_STOPBIT);

    if ((_receive_buffer_tail + 1) % _SS_MAX_RX_BUFF != _receive_buffer_head) 
    {
      // save new data in buffer: tail points to where byte goes
      _receive_buffer[_receive_buffer_tail] = d; // save new byte
      _receive_buffer_tail = (_receive_buffer_tail + 1) % _SS_MAX_RX_BUFF;
    } 
  }
}

void SoftwareSerial_tx_pin_write(uint8_t pin_state)
{
  if (pin_state == LOW)
    *_transmitPortRegister &= ~_transmitBitMask;
  else
    *_transmitPortRegister |= _transmitBitMask;
}

uint8_t SoftwareSerial_rx_pin_read()
{
  return *_receivePortRegister & _receiveBitMask;
}

ISR(PCINT0_vect)
{
  SoftwareSerial_recv();
}

ISR(PCINT1_vect)
{
  SoftwareSerial_recv();
}

ISR(PCINT2_vect)
{
  SoftwareSerial_recv();
}

void SoftwareSerial_setTX(uint8_t tx)
{
  pinMode(tx, OUTPUT);
  digitalWrite(tx, HIGH);
  _transmitBitMask = digitalPinToBitMask(tx);
  uint8_t port = digitalPinToPort(tx);
  _transmitPortRegister = portOutputRegister(port);
}

void SoftwareSerial_setRX(uint8_t rx)
{
  pinMode(rx, INPUT);
  digitalWrite(rx, HIGH);  // pullup for normal logic!
  _receivePin = rx;
  _receiveBitMask = digitalPinToBitMask(rx);
  uint8_t port = digitalPinToPort(rx);
  _receivePortRegister = portInputRegister(port);
}

void SoftwareSerial_end()
{
  if (digitalPinToPCMSK(_receivePin))
    *digitalPinToPCMSK(_receivePin) &= ~_BV(digitalPinToPCMSKbit(_receivePin));
}

int SoftwareSerial_read()
{
  // Empty buffer?
  if (_receive_buffer_head == _receive_buffer_tail)
    return -1;

  // Read from "head"
  uint8_t d = _receive_buffer[_receive_buffer_head]; // grab next byte
  _receive_buffer_head = (_receive_buffer_head + 1) % _SS_MAX_RX_BUFF;
  return d;
}

int SoftwareSerial_available()
{
  return (_receive_buffer_tail + _SS_MAX_RX_BUFF - _receive_buffer_head) % _SS_MAX_RX_BUFF;
}

size_t SoftwareSerial_write(uint8_t b)
{
  uint8_t oldSREG = SREG;
  cli();  // turn off interrupts for a clean txmit

  // Write the start bit
  SoftwareSerial_tx_pin_write(LOW);
  SoftwareSerial_tunedDelay(TX_DELAY + XMIT_START_ADJUSTMENT);

  // Write each of the 8 bits
  for (byte mask = 0x01; mask; mask <<= 1)
  {
    if (b & mask) // choose bit
      SoftwareSerial_tx_pin_write(HIGH); // send 1
    else
      SoftwareSerial_tx_pin_write(LOW); // send 0
  
    SoftwareSerial_tunedDelay(TX_DELAY);
  }
  SoftwareSerial_tx_pin_write(HIGH); // restore pin to natural state

  SREG = oldSREG; // turn interrupts back on
  SoftwareSerial_tunedDelay(TX_DELAY);
  
  return 1;
}

int SoftwareSerial_peek()
{
  // Empty buffer?
  if (_receive_buffer_head == _receive_buffer_tail)
    return -1;

  // Read from "head"
  return _receive_buffer[_receive_buffer_head];
}

void SoftwareSerial_flush()
{
  uint8_t oldSREG = SREG;
  cli();
  _receive_buffer_head = _receive_buffer_tail = 0;
  SREG = oldSREG;
}
#endif // ESP

