// Nodo Compatible Wireless Word Clock V0.01
// 23-12-2014
// Project state: Proof of Concept

#define UNIT_NODO                          28
#define EVENT_PAR1                        123

// select communication type (enable only one !)
#define NRF
//#define ESP

// In case of ESP module, set wifi settings here:
#define ESP_SSID       "<SSID>"
#define ESP_PASS       "<password"
#define ESP_BROADCAST   "192.168.0.255"

#define DEBUG

//Library includes
#include <FastLED.h>
#include <Wire.h>
#include <Time.h>
#include <DS1307RTC.h>
#include <IRremote.h>

// IR defines 
#define ON 0xFFE01F 
#define OFF 0xFF609F
#define DIM 0xFF20DF
#define BRIGHTER 0xFFA05F
#define RED 0xFF906F
#define GREEN 0xFF10EF
#define BLUE 0xFF50AF
#define WHITE 0xFFD02F
#define FAST 0xFFF00F //fast Test, button Flash
#define FASTA 0xFF708F //fast Test2, button light blue
#define PARTY 0xFFE817 //make Party, button Strobe
#define BRIGHT 0xFFD827 //autobrightness enable, button Fade
#define HEART 0xFFC837 //show Heart, button Smooth
#define CLOCK 0xFF8877//clock Logic, button yellow --> left under
#define NEXT 0xFFB04F //next Color, button orange --> left up
#define PREV 0xFFA857 //previous Color, button orange 2 --> left

//LED defines
#define NUM_LEDS 114

//PIN defines
#define STRIP_DATA_PIN 6
#define IR_RECV_PIN 11
#define ARDUINO_LED 13 //Default Arduino LED
#define LDR_PIN A0

uint8_t strip[NUM_LEDS];
uint8_t stackptr = 0;

CRGB leds[NUM_LEDS];

IRrecv irrecv = IRrecv(IR_RECV_PIN);
decode_results irDecodeResults;

boolean autoBrightnessEnabled = true;

int displayMode = CLOCK;

CRGB defaultColor = CRGB::White;
uint8_t colorIndex = 0;

int testHours = 0;
int testMinutes = 0;

//multitasking helper
const long oneSecondDelay = 1000;
const long halfSecondDelay = 500;

long waitUntilRtc = 0;
long waitUntilParty = 0;
long waitUntilOff = 0;
long waitUntilFastTest = 0;
long waitUntilFastTesta = 0;
long waitUntilHeart = 0;
long waitUntilLDR = 0;


//forward declaration
void fastTest();
void fastTesta();
void clockLogic();
void doIRLogic();
void doLDRLogic();
void makeParty();
void showHeart();
void pushToStrip(int ledId);
void resetAndBlack();
void resetStrip();
void displayStripRandomColor();
void displayStrip();
void displayStrip(CRGB colorCode);
void timeToStrip(uint8_t hours,uint8_t minutes);

void setup() {
#ifdef DEBUG
  Serial.begin(9600);
#endif

  //setup pinmode
  pinMode(ARDUINO_LED, OUTPUT);
  pinMode(LDR_PIN, INPUT);	
  pinMode(2, INPUT);       // set pin 2 as an input
  digitalWrite(2, HIGH);   // activate internal pullup resistor on pin 2
  pinMode(3, INPUT);       // set pin 3 as an input
  digitalWrite(3, HIGH);   // activate internal pullup resistor on pin 3
  pinMode(4, INPUT);       // set pin 4 as an input
  digitalWrite(4, HIGH);   // activate internal pullup resistor on pin 4

  //setup leds incl. fastled
  for(int i = 0; i<NUM_LEDS; i++) {
    strip[i] = 0;
  }
  FastLED.addLeds<WS2812B, STRIP_DATA_PIN, GRB>(leds, NUM_LEDS);
  resetAndBlack();
  displayStrip();

  //setup rtc
  setSyncProvider(RTC.get);   // the function to get the time from the RTC
  setSyncInterval(60);        // set the number of seconds between re-sync

  if(timeStatus()!= timeSet)
    Serial.println("Unable to sync with the RTC.");
  else
    Serial.println("RTC has been read.");

  //setup ir
  irrecv.enableIRIn();
#ifdef ESP
  esp_init();
#endif

#ifdef NRF
  nrf_init();
#endif

}
void loop(){

  byte cmd=0;
  
#ifdef ESP
  cmd=esp_event();
#endif

#ifdef NRF
  cmd=nrf_event();
#endif

  switch(cmd)
  {
  case 1:
    displayMode = ON;
    break;
  case 2:
    displayMode = OFF;
    break;
  case 3:
    displayMode = DIM;
    break;
  case 4:
    displayMode = BRIGHTER;
    break;
  case 5:
    displayMode = RED;
    break;
  case 6:
    displayMode = GREEN;
    break;
  case 7:
    displayMode = BLUE;
    break;
  case 8:
    displayMode = WHITE;
    break;
  case 9:
    displayMode = FAST;
    break;
  case 10:
    displayMode = FASTA;
    break;
  case 11:
    displayMode = PARTY;
    break;
  case 12:
    displayMode = BRIGHT;
    break;
  case 13:
    displayMode = HEART;
    break;
  case 14:
    displayMode = CLOCK;
    testMinutes = -1;
    testHours = -1;
    break;
  case 15:
    displayMode = PREV;
    break;
  case 16:
    displayMode = NEXT;
    break;
  }

  if (Serial.available())
  {
    char c=Serial.read();
    if (c == 'p')
    {
      displayMode = PARTY;
    }
    if (c == 'c')
    {
      displayMode = CLOCK;
      autoBrightnessEnabled = true;
      //to force display update
      testMinutes = -1;
      testHours = -1;
    }
    if (c == 'h')
    {
      displayMode = HEART;
    }
    if (c == 'g')
    {
      displayMode = GREEN;
    }
    if (c == 'r')
    {
      displayMode = RED;
    }
    if (c == 'b')
    {
      displayMode = BLUE;
    }
    if (c == 'w')
    {
      displayMode = WHITE;
    }
    if (c == '+')
    {
      setTime(now()+3600);                  // adjust onboard time by +3600 seconds (1hr) ***
      RTC.set(now());                    // sending the onboard time to the RTC.
      Serial.println("Added an hour.");  // send confirmation to the PC.
      Serial.println(now());
    }

    if (c == '-')
    {
      setTime(now()-3600);                  // adjust onboard time by +3600 seconds (1hr) ***
      RTC.set(now());                    // sending the onboard time to the RTC.
      Serial.println("Minus one hour.");  // send confirmation to the PC.
      Serial.println(now());
    }

    if (c == '>')
    {
      setTime(now()+60);                  // adjust onboard time by +3600 seconds (1hr) ***
      RTC.set(now());                    // sending the onboard time to the RTC.
      Serial.println("Added one Minute.");  // send confirmation to the PC.
      Serial.println(now());
    }

    if (c == '<')
    {
      setTime(now()-60);                  // adjust onboard time by +3600 seconds (1hr) ***
      RTC.set(now());                    // sending the onboard time to the RTC.
      Serial.println("Minus one minute.");  // send confirmation to the PC.
      Serial.println(now());
    }
    Serial.flush();
  }

  //time setting with buttons	
  if (digitalRead(2) == LOW) {         // Zero the minutes!
    setTime(hour(),0,second(),day(),month(), year());
    RTC.set(now());                    // sending the onboard time to the RTC.
    Serial.println("Minutes Reset.");  // send confirmation to the PC.
  }
  if (digitalRead(3) == LOW) {         // add an hour!
    setTime(now()+3600);               // adjust onboard time by +3600 seconds (1hr) ***
    RTC.set(now());                    // sending the onboard time to the RTC.
    Serial.println("Added an hour.");  // send confirmation to the PC.
    delay(900);
  }
  if (digitalRead(4) == LOW) {         // add a minute!
    setTime(now()+60);                 // adjust onboard time by +60 seconds (1min) ***
    RTC.set(now());                    // sending the onboard time to the RTC.
    Serial.println("Added a minute."); // send confirmation to the PC.
    delay(900);
  }

  delay(100);            // wait for a bit.

  doIRLogic();
  doLDRLogic();
  
  switch(displayMode) {
  case ON: //divided into On and Off
    clockLogic();
    break;
  case OFF: //divided into On and Off
    off();
    break;
  case RED:
    clockLogic();
    displayStrip(CRGB::Red);
    break;
  case GREEN:
    clockLogic();
    displayStrip(CRGB::Green);
    break;
  case BLUE:
    clockLogic();
    displayStrip(CRGB::Blue);
    break;
  case WHITE:
    clockLogic();
    displayStrip(CRGB::White);
    break;
  case FAST:
    fastTest();
    break;		
  case FASTA:
    fastTesta();
    break;
  case PARTY:
    makeParty();
    break;
  case HEART:
    showHeart();
    break;
  case CLOCK:
    clockLogic();
    break;
  default:
    clockLogic();
    break;
  }
}
void digitalClockDisplay(){
  // digital clock display of the time 
  Serial.print(hour());
  printDigits(minute());
  printDigits(second());			
  Serial.println(); 
} 
void printDigits(int digits){
  // utility function for digital clock display: prints preceding colon and leading 0
  Serial.print(":");
  if(digits < 10)
    Serial.print('0');
  Serial.print(digits);
}
void doLDRLogic() {
  if(millis() >= waitUntilLDR && autoBrightnessEnabled) {
    waitUntilLDR = millis();
    int ldrVal = map(analogRead(LDR_PIN), 0, 1023, 15, 245);
    FastLED.setBrightness(255-ldrVal);
    FastLED.show();
    waitUntilLDR += oneSecondDelay;
  }
}
void doIRLogic() {
  uint8_t brightness = 0;
  if (irrecv.decode(&irDecodeResults)) {
    delay(50);
    switch(irDecodeResults.value) {
    case ON: // divided into On and Off
      displayMode = ON;
      autoBrightnessEnabled = true;
      //to force display update
      testMinutes = -1;
      testHours = -1;
      break;
    case OFF:
      OFF;
      displayMode = OFF;
      break;
    case DIM:
      autoBrightnessEnabled = false;
      brightness = FastLED.getBrightness();
      if(brightness >= 40) {
        FastLED.setBrightness(brightness - 35);
      } 
      else {
        FastLED.setBrightness(0);
      }
      FastLED.show();
      break;
    case BRIGHTER:
      autoBrightnessEnabled = false;
      brightness = FastLED.getBrightness();
      if(brightness <= 255 - 35) {
        FastLED.setBrightness(brightness + 35);
      } 
      else {
        FastLED.setBrightness(255);
      }
      FastLED.show();
      break;
    case RED:
      displayMode = RED;
      break;
    case GREEN:
      displayMode = GREEN;
      break;
    case BLUE:
      displayMode = BLUE;
      break;
    case WHITE:
      displayMode = WHITE;
      break;
    case FAST:
      displayMode = FAST;
      break;	
    case FASTA:
      displayMode = FASTA;
      break;	
    case PARTY:
      displayMode = PARTY;
      break;	
    case BRIGHT:
      displayMode = ON;
      autoBrightnessEnabled = true;
      //to force display update
      testMinutes = -1;
      testHours = -1;
      break;	
    case HEART:
      displayMode = HEART;
      break;	
    case CLOCK:
      displayMode = CLOCK;
      autoBrightnessEnabled = true;
      //to force display update
      testMinutes = -1;
      testHours = -1;
      break;
    case NEXT:
      defaultColor = nextColor();
      displayStrip();
      break;
    case PREV:
      defaultColor = prevColor();
      displayStrip();
      break;
    default:
      break;
    }
    irrecv.resume();
  }
}
///////////////////////
//DISPLAY MODES
///////////////////////
void clockLogic() {
  if(millis() >= waitUntilRtc) {
    waitUntilRtc = millis();
    if(testMinutes != minute() || testHours != hour()) {
      testMinutes = minute();
      testHours = hour();
      resetAndBlack();
      timeToStrip(testHours, testMinutes);
      displayStrip(defaultColor);
    }
    waitUntilRtc += oneSecondDelay;
  }
}
void off() {
  if(millis() >= waitUntilOff) {
    waitUntilOff = millis();
    resetAndBlack();
    displayStrip(CRGB::Black);
    waitUntilOff += halfSecondDelay;
  }
}
void makeParty() {
  if(millis() >= waitUntilParty) {
    autoBrightnessEnabled = false;
    waitUntilParty = millis();
    resetAndBlack();
    for(int i = 0; i<NUM_LEDS;i++) {
      leds[i] = CHSV(random(0, 255), 255, 255);
    }
    FastLED.show();
    waitUntilParty += halfSecondDelay;
  }
}
void showHeart() {
  if(millis() >= waitUntilHeart) {
    autoBrightnessEnabled = false;
    waitUntilHeart = millis();
    resetAndBlack();
    pushToStrip(3); 
    pushToStrip(22); 
    pushToStrip(24); 
    pushToStrip(45); //posities veranderd
    pushToStrip(46); 
    pushToStrip(66); 
    pushToStrip(70); 
    pushToStrip(86);
    pushToStrip(94); 
    pushToStrip(107); 
    pushToStrip(96);
    pushToStrip(82); 
    pushToStrip(76);
    pushToStrip(58); 
    pushToStrip(56);
    pushToStrip(35); 
    pushToStrip(34);
    pushToStrip(14); 
    pushToStrip(9);
    pushToStrip(8); 
    pushToStrip(17);
    pushToStrip(29); 
    pushToStrip(9);
    pushToStrip(4); 
    pushToStrip(19);
    displayStrip(CRGB::Red);
    waitUntilHeart += oneSecondDelay;
  }
}

void fastTesta() {
  if(millis() >= waitUntilFastTesta) {
    autoBrightnessEnabled = false;
    waitUntilFastTesta = millis();
    resetAndBlack();
    // test all leds white 
    for(int led = 0; led < 115; led++) {
      leds[led] = CRGB::White; 
    }
    // Show the leds
    FastLED.show();

    // Turn our current leds back to black 
    resetAndBlack();
    waitUntilFastTesta += halfSecondDelay;
  }
}
void fastTest() {
  if(millis() >= waitUntilFastTest) {
    autoBrightnessEnabled = false;
    waitUntilFastTest = millis();
    if(testMinutes >= 60) {
      testMinutes = 0;
      testHours++;
    }
    if(testHours >= 24) {
      testHours = 0;
    }

    //Array leeren
    resetAndBlack();
    timeToStrip(testHours, testMinutes);
    displayStripRandomColor();
    testMinutes++;
    waitUntilFastTest += oneSecondDelay;
  }
}

CRGB prevColor() {
  if(colorIndex > 0) {
    colorIndex--;
  }
  return getColorForIndex();
}
CRGB nextColor() {
  if(colorIndex < 9) {
    colorIndex++;
  }
  return getColorForIndex();
}
CRGB getColorForIndex() {
  switch(colorIndex) {
  case 0:
    return CRGB::White;
  case 1:
    return CRGB::Blue;
  case 2:
    return CRGB::Aqua;
  case 3:
    return CRGB::Green;
  case 4:
    return CRGB::Lime;
  case 5:
    return CRGB::Red;
  case 6:
    return CRGB::Magenta;
  case 7:
    return CRGB::Olive;
  case 8:
    return CRGB::Yellow;
  case 9:
    return CRGB::Silver;
  default:
    colorIndex = 0;
    return CRGB::White;
  }
}
void pushToStrip(int ledId) {
  strip[stackptr] = ledId;
  stackptr++;
}
void resetAndBlack() {
  resetStrip();
  for(int i = 0; i<NUM_LEDS; i++) {
    leds[i] = CRGB::Black;
  }
}
void resetStrip() {
  stackptr = 0;
  for(int i = 0; i<NUM_LEDS; i++) {
    strip[i] = 0;
  }
}
void displayStripRandomColor() {
  for(int i = 0; i<stackptr; i++) {
    leds[strip[i]] = CHSV(random(0, 255), 255, 255);
  }
  FastLED.show();
}
void displayStrip() {
  displayStrip(defaultColor);
}
void displayStrip(CRGB colorCode) {
  for(int i = 0; i<stackptr; i++) {
    leds[strip[i]] = colorCode;
  }
  FastLED.show();
}

void timeToStrip(uint8_t hours,uint8_t minutes)
{
  pushIT_IS();
  //show minutes
  if(minutes >= 5 && minutes < 10) {
    pushFIVE1();
    pushAFTER();
  } 
  else if(minutes >= 10 && minutes < 15) {
    pushTEN1();
    pushAFTER();
  } 
  else if(minutes >= 15 && minutes < 20) {
    pushQUATER();
    pushAFTER();
  } 
  else if(minutes >= 20 && minutes < 25) {
    pushTEN1();
    pushFOR();
    pushHALF();
  } 
  else if(minutes >= 25 && minutes < 30) {
    pushFIVE1();
    pushFOR();
    pushHALF();
  } 
  else if(minutes >= 30 && minutes < 35) {
    pushHALF();
  } 
  else if(minutes >= 35 && minutes < 40) {
    pushFIVE1();
    pushAFTER();
    pushHALF();
  } 
  else if(minutes >= 40 && minutes < 45) {
    pushTEN1();
    pushAFTER();
    pushHALF();
    //}
  } 
  else if(minutes >= 45 && minutes < 50) {
    pushQUATER();
    pushFOR();
  } 
  else if(minutes >= 50 && minutes < 55) {
    pushTEN1();
    pushFOR();
  } 
  else if(minutes >= 55 && minutes < 60) {
    pushFIVE1();
    pushFOR();
  }

  int singleMinutes = minutes % 5;
  switch(singleMinutes) {
  case 1:
    pushM_ONE();
    //                       DEBUG_PRINT("led 0 aan");
    break;
  case 2:
    pushM_ONE();
    pushM_TWO();
    break;
  case 3:
    pushM_ONE();
    pushM_TWO();
    pushM_THREE();
    break;
  case 4:
    pushM_ONE();
    pushM_TWO();
    pushM_THREE();
    pushM_FOUR();
    break;
  }
  if(hours >= 12) {
    hours -= 12;
  }
  if(hours == 12) {
    hours = 0;
  }
  if(minutes >= 20) {
    hours++;
  }

  //show hours
  switch(hours) {
  case 0:
    pushTWELVE();
    break;
  case 1:
    pushONE();
    break;
  case 2:
    pushTWO();
    break;
  case 3:
    pushTHREE();
    break;
  case 4:
    pushFOUR();
    break;
  case 5:
    pushFIVE2();
    break;
  case 6:
    pushSIX();
    break;
  case 7:
    pushSEVEN();
    break;
  case 8:
    pushEIGHT();
    break;
  case 9:
    pushNINE();
    break;
  case 10:
    pushTEN();
    break;
  case 11:
    pushELEVEN();
    break;
  case 12:
    pushTWELVE();
    break;
  }
  //show HOURE
  if(minutes < 5) {
    pushHOURE();
  }
}

///////////////////////
//PUSH WORD HELPER
///////////////////////
void pushM_ONE() {
  pushToStrip(0);
}
void pushM_TWO() {
  pushToStrip(12);
}
void pushM_THREE() {
  pushToStrip(101);
}
void pushM_FOUR() {
  pushToStrip(113);
}
void pushIT_IS()  {
  pushToStrip(1);
  pushToStrip(2);
  pushToStrip(3);
  pushToStrip(5);
  pushToStrip(6);
}
void pushAFTER() {
  pushToStrip(36);
  pushToStrip(37);
  pushToStrip(38);
  pushToStrip(39);
}
void pushQUATER() {
  pushToStrip(30);
  pushToStrip(31);
  pushToStrip(32);
  pushToStrip(33);
  pushToStrip(34);
}
void pushFOR() {
  pushToStrip(41);
  pushToStrip(42);
  pushToStrip(43);
  pushToStrip(44);
}
void pushHALF() {
  pushToStrip(50);
  pushToStrip(51);
  pushToStrip(52);
  pushToStrip(53);
}
void pushONE()  {
  pushToStrip(63);
  pushToStrip(64);
  pushToStrip(65);
}
void pushTWO() {
  pushToStrip(64);
  pushToStrip(65);
  pushToStrip(66);
  pushToStrip(67);
}
void pushTHREE() {
  pushToStrip(109);
  pushToStrip(110);
  pushToStrip(111);
  pushToStrip(112);
}
void pushFOUR() {
  pushToStrip(57);
  pushToStrip(58);
  pushToStrip(59);
  pushToStrip(60);
}
void pushFIVE1() {
  pushToStrip(8);
  pushToStrip(9);
  pushToStrip(10);
  pushToStrip(11);
}
void pushFIVE2() {
  pushToStrip(92);
  pushToStrip(93);
  pushToStrip(94);
  pushToStrip(95);
}
void pushSIX() {
  pushToStrip(69);
  pushToStrip(88);
  pushToStrip(91);
}
void pushSEVEN() {
  pushToStrip(69);
  pushToStrip(70);
  pushToStrip(71);
  pushToStrip(72);
  pushToStrip(73);
}
void pushEIGHT() {
  pushToStrip(97);
  pushToStrip(98);
  pushToStrip(99);
  pushToStrip(100);
}
void pushNINE() {
  pushToStrip(73);
  pushToStrip(74);
  pushToStrip(75);
  pushToStrip(76);
  pushToStrip(77);
}
void pushTEN() {
  pushToStrip(54);
  pushToStrip(59);
  pushToStrip(76);
  pushToStrip(81);
}
void pushTEN1() {
  pushToStrip(25);
  pushToStrip(26);
  pushToStrip(27);
  pushToStrip(28);
}
void pushELEVEN() {
  pushToStrip(107);
  pushToStrip(108);
  pushToStrip(109);
}
void pushTWELVE() {
  pushToStrip(82);
  pushToStrip(83);
  pushToStrip(84);
  pushToStrip(85);
  pushToStrip(86);
  pushToStrip(87);
}
void pushTWENTY() {
  pushToStrip(16);
  pushToStrip(17);
  pushToStrip(18);
  pushToStrip(19);
  pushToStrip(20);
  pushToStrip(21);
  pushToStrip(22);
}
void pushHOURE() {
  pushToStrip(102);
  pushToStrip(103);
  pushToStrip(104);
}


